from __future__ import annotations

__all__ = ["RasrFsaBuilder", "WeightedFsa", "RasrFsaBuilderV2", "WeightedFsaV2"]

from functools import reduce
from typing import Iterable, NamedTuple, Tuple, Union

import numpy as np
import torch


class WeightedFsa(NamedTuple):
    """
    Convenience class that represents an FSA. It supports scaling the weights of the
    fsa by simple left-multiplication and moving the tensors to a different device.
    It can simply be passed to :func:`i6_native_ops.fbw.fbw_loss` and :func:`i6_native_ops.fast_viterbi.align_viterbi`.

    :param num_states: the total number of all states S
    :param edges: a [4, E] tensor of edges with number of edges E and where each column is an edge
        consisting of from-state, to-state, emission idx and the index of the sequence it belongs to
    :param weights: a [E,] tensor of weights for each edge scaled by the tdp_scale
    :param start_end_states: a [N, 2] tensor of start and end states for each of the N sequences
    """

    num_states: torch.IntTensor
    edges: torch.IntTensor
    weights: torch.FloatTensor
    start_end_states: torch.IntTensor

    def __mul__(self, scale: float) -> WeightedFsa:
        """Multiply the weights, i.e. the third element, with a scale."""
        return WeightedFsa(
            self.num_states,
            self.edges,
            self.weights * scale,
            self.start_end_states,
        )

    def to(self, device: Union[str, torch.device]) -> WeightedFsa:
        """Move the tensors to a given device. This wraps around the
        PyTorch `Tensor.to(device)` method."""
        return WeightedFsa._make(tensor.to(device) for tensor in self)


class WeightedFsaV2(NamedTuple):
    """
    Convenience class that represents an FSA for the fbw2 loss from i6_native_ops. It supports scaling the weights of
    the FSA by simple left-multiplication and moving the tensors to a different device.
    It can be passed to :func:`i6_native_ops.fbw2.fbw2_loss`.

    :param num_states: the number of states per sequence (shape: [N])
    :param num_edges: the number of edges per sequence (shape: [N])
    :param edges: a [3, E] tensor of edges with number of edges E and where each column is an edge
        consisting of from-state, to-state and emission idx
    :param weights: a [E,] tensor of weights for each edge scaled by the tdp_scale
    :param start_end_states: a [N, 2] tensor of start and end states for each of the N sequences
    """

    num_states: torch.IntTensor
    num_edges: torch.IntTensor
    edges: torch.IntTensor
    weights: torch.FloatTensor
    start_end_states: torch.IntTensor

    def __mul__(self, scale: float) -> WeightedFsaV2:
        """Multiply the weights, i.e. the forth element, with a scale."""
        return WeightedFsaV2(
            self.num_states,
            self.num_edges,
            self.edges,
            self.weights * scale,
            self.start_end_states,
        )

    def to(self, device: Union[str, torch.device]) -> WeightedFsaV2:
        """
        Move the tensors that can be on device to a given device.
        Some Tensors (num_states/num_edges) are expected to be in CPU memory.
        """

        return WeightedFsaV2(
            self.num_states.to("cpu"),
            self.num_edges.to("cpu"),
            self.edges.to(device),
            self.weights.to(device),
            self.start_end_states.to(device),
        )


class RasrFsaBuilder:
    """
    Builder class that wraps around the librasr.AllophoneStateFsaBuilder,
    bringing the FSAs into the correct format for the `i6_native_ops.fbw.fbw_loss`.
    Use of this class requires a working installation of the python package `librasr`.
    Hence, the package is locally imported in case other classes are accessed from
    this module.
    This class provides an explicit implementation of the `__getstate__` and `__setstate__`
    functions, necessary for pickling as the C++-class `librasr.AllophoneStateFsaBuilder`
    is not picklable.

    :param config_path: path to the RASR fsa exporter config
    :param tdp_scale: multiply the weights by this scale
    """

    def __init__(self, config_path: str, tdp_scale: float = 1.0):
        import librasr

        self.config_path = config_path
        config = librasr.Configuration()
        config.set_from_file(self.config_path)
        self.builder = librasr.AllophoneStateFsaBuilder(config)
        self.tdp_scale = tdp_scale

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["builder"]
        return state

    def __setstate__(self, state):
        import librasr

        self.__dict__.update(state)
        config = librasr.Configuration()
        config.set_from_file(self.config_path)
        self.builder = librasr.AllophoneStateFsaBuilder(config)

    def build_single(self, seq_tag: str) -> Tuple[int, int, np.ndarray, np.ndarray]:
        """
        Build the FSA for the given sequence tag in the corpus.

        :param seq_tag: sequence tag
        :return: FSA as a tuple containing
            * number of states S
            * number of edges E
            * integer edge array of shape [E, 3] where each row is an edge
                consisting of from-state, to-state and the emission idx
            * float weight array of shape [E,]
        """
        raw_fsa = self.builder.build_by_segment_name(seq_tag)
        return raw_fsa

    def build_batch(self, seq_tags: Iterable[str]) -> WeightedFsa:
        """
        Build and concatenate the FSAs for a batch of sequence tags
        and reformat as an input to `i6_native_ops.fbw.fbw_loss`.
        Here the FSAs are concatenated to a long FSA with multiple start and
        end states corresponding to each single FSA. For the concatenation,
        the state IDs of each single FSA are incrememented and made unique in
        the batch.
        Additionally we apply an optional scale to the weights.

        :param seq_tags: an iterable object of sequence tags
        :return: a concatenated FSA
        """

        def append_fsa(a, b):
            edges = torch.from_numpy(np.int32(b[2])).reshape((3, b[1]))
            return (
                a[0] + [b[0]],  # num states
                a[1] + [b[1]],  # num edges
                torch.hstack([a[2], edges]),  # edges
                torch.cat([a[3], torch.from_numpy(b[3])]),  # weights
            )

        # concatenate all FSAs in the batch into a single one where state ids are not yet unique
        fsas = map(self.build_single, seq_tags)
        empty_fsa = ([], [], torch.empty((3, 0), dtype=torch.int32), torch.empty((0,)))
        num_states, num_edges, all_edges, all_weights = reduce(append_fsa, fsas, empty_fsa)
        num_edges = torch.tensor(num_edges, dtype=torch.int32)
        num_states = torch.tensor(num_states, dtype=torch.int32)

        # accumulate number of states for each single fsa in order to determine start and end states
        # and make states in edge tensor unique to each sequence
        cum_num_states = torch.cumsum(num_states, dim=0, dtype=torch.int32)
        state_offsets = torch.cat([torch.zeros((1,), dtype=torch.int32), cum_num_states[:-1]])
        start_end_states = torch.vstack([state_offsets, cum_num_states - 1])

        # add unique sequence ids to the edge tensor and add start states to the states
        # in order to make them unique
        edge_seq_idxs = torch.repeat_interleave(num_edges)
        all_edges[:2, :] += torch.repeat_interleave(state_offsets, num_edges)
        all_edges = torch.vstack([all_edges, edge_seq_idxs])

        out_fsa = WeightedFsa(
            cum_num_states[-1],
            all_edges,
            all_weights,
            start_end_states,
        )

        if self.tdp_scale != 1.0:
            out_fsa *= self.tdp_scale

        return out_fsa


class RasrFsaBuilderV2(RasrFsaBuilder):
    """
    An update of the RasrFsaBuilder that is compatible with the fbw2 op from i6_native_ops.
    """

    def build_batch(self, seq_tags: Iterable[str]) -> WeightedFsaV2:
        fsas = list(map(self.build_single, seq_tags))

        num_states = [f[0] for f in fsas]
        num_edges = [f[1] for f in fsas]
        start_states = np.cumsum([0] + num_states)[:-1]
        end_states = np.cumsum(num_states) - 1
        weights = np.concatenate(tuple(f[3] for f in fsas))

        edges = []
        for idx, f in enumerate(fsas):
            f_edges = f[2].reshape(3, -1).copy()
            f_edges[:2, :] += start_states[idx]
            edges.append(f_edges)

        out_fsa = WeightedFsaV2(
            torch.IntTensor(num_states).to(torch.uint32),
            torch.IntTensor(num_edges).to(torch.uint32),
            torch.IntTensor(np.concatenate(edges, axis=1)).contiguous(),
            torch.Tensor(weights),
            torch.IntTensor(np.array([start_states, end_states])),
        )

        if self.tdp_scale != 1.0:
            out_fsa *= self.tdp_scale

        return out_fsa



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.linalg import vector_norm

__all__ = [
    "RandomProjectionQuantizer",
]


class RandomProjectionQuantizer(nn.Module):
    """
    implement the fixed random projection quantizer from BestRQ
    C.f. https://arxiv.org/pdf/2202.01855 for theoretic background
    code adapted from https://github.com/speechbrain/speechbrain/blob/16b6420d4ff23210cfca2e888be8853264e0cb17/speechbrain/nnet/quantisers.py#L127
    """

    def __init__(self, input_dim, codebook_dim, codebook_num_vars):
        """
        :param input_dim: number of feature dimension of input
        :param codebook_dim: number of dimension for vocab in the codebook
        :param codebook_num_vars: vocab size of the codebook
        """
        super().__init__()

        self.input_dim = input_dim

        # projection matrix use Xavier initialization
        P_init = torch.empty((input_dim, codebook_dim))
        self.register_buffer("P", nn.init.xavier_uniform_(P_init))

        # normalize random matrix for codebook
        self.register_buffer("CB", F.normalize(torch.randn(codebook_num_vars, codebook_dim)))

    def forward(self, x: torch.tensor) -> torch.tensor:
        x = F.normalize(x @ self.P)
        return vector_norm((self.CB.unsqueeze(1) - x.unsqueeze(1)), dim=-1).argmin(dim=1)



from __future__ import annotations

__all__ = [
    "ConformerRelPosBlockV1Config",
    "ConformerRelPosEncoderV1Config",
    "ConformerRelPosBlockV1",
    "ConformerRelPosEncoderV1",
]

import torch
from torch import nn
from dataclasses import dataclass, field
from typing import List

from i6_models.config import ModelConfiguration, ModuleFactoryV1
from i6_models.parts.conformer import (
    ConformerConvolutionV2,
    ConformerConvolutionV2Config,
    ConformerMHSARelPosV1,
    ConformerMHSARelPosV1Config,
    ConformerPositionwiseFeedForwardV2,
    ConformerPositionwiseFeedForwardV2Config,
)
from i6_models.assemblies.conformer import ConformerEncoderV2


@dataclass
class ConformerRelPosBlockV1Config(ModelConfiguration):
    """
    Attributes:
        ff_cfg: Configuration for ConformerPositionwiseFeedForwardV2
        mhsa_cfg: Configuration for ConformerMHSARelPosV1
        conv_cfg: Configuration for ConformerConvolutionV2
        modules: List of modules to use for ConformerRelPosBlockV1,
            "ff" for feed forward module, "mhsa" for multi-head self attention module, "conv" for conv module
        scales: List of scales to apply to the module outputs before the residual connection
    """

    # nested configurations
    ff_cfg: ConformerPositionwiseFeedForwardV2Config
    mhsa_cfg: ConformerMHSARelPosV1Config
    conv_cfg: ConformerConvolutionV2Config
    modules: List[str] = field(default_factory=lambda: ["ff", "mhsa", "conv", "ff"])
    scales: List[float] = field(default_factory=lambda: [0.5, 1.0, 1.0, 0.5])

    def __post__init__(self):
        super().__post_init__()
        assert len(self.modules) == len(self.scales), "modules and scales must have same length"
        for module_name in self.modules:
            assert module_name in ["ff", "mhsa", "conv"], "module not supported"


class ConformerRelPosBlockV1(nn.Module):
    """
    Conformer block module, modifications compared to ConformerBlockV1:
    - uses ConfomerMHSARelPosV1 as MHSA module
    - enable constructing the block with self-defined module_list as ConformerBlockV2
    """

    def __init__(self, cfg: ConformerRelPosBlockV1Config):
        """
        :param cfg: conformer block configuration with subunits for the different conformer parts
        """
        super().__init__()

        modules = []
        for module_name in cfg.modules:
            if module_name == "ff":
                modules.append(ConformerPositionwiseFeedForwardV2(cfg=cfg.ff_cfg))
            elif module_name == "mhsa":
                modules.append(ConformerMHSARelPosV1(cfg=cfg.mhsa_cfg))
            elif module_name == "conv":
                modules.append(ConformerConvolutionV2(model_cfg=cfg.conv_cfg))
            else:
                raise NotImplementedError

        self.module_list = nn.ModuleList(modules)
        self.scales = cfg.scales
        self.final_layer_norm = torch.nn.LayerNorm(cfg.ff_cfg.input_dim)

    def forward(self, x: torch.Tensor, /, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        :param x: input tensor of shape [B, T, F]
        :param sequence_mask: mask tensor where 1 defines positions within the sequence and 0 outside, shape: [B, T]
        :return: torch.Tensor of shape [B, T, F]
        """
        for scale, module in zip(self.scales, self.module_list):
            if isinstance(module, ConformerMHSARelPosV1):
                x = scale * module(x, sequence_mask) + x
            else:
                x = scale * module(x) + x
        x = self.final_layer_norm(x)  #  [B, T, F]
        return x


@dataclass
class ConformerRelPosEncoderV1Config(ModelConfiguration):
    """
    Attributes:
        num_layers: Number of conformer layers in the conformer encoder
        frontend: A pair of ConformerFrontend and corresponding config
        block_cfg: Configuration for ConformerRelPosBlockV1
    """

    num_layers: int

    # nested configurations
    frontend: ModuleFactoryV1
    block_cfg: ConformerRelPosBlockV1Config


class ConformerRelPosEncoderV1(ConformerEncoderV2):
    """
    Modifications compared to ConformerEncoderV2:
    - supports Shaw's relative positional encoding using learnable position embeddings
      and Transformer-XL style relative PE using fixed sinusoidal or learnable position embeddings
    """

    def __init__(self, cfg: ConformerRelPosEncoderV1Config):
        """
        :param cfg: conformer encoder configuration with subunits for frontend and conformer blocks
        """
        super().__init__(cfg)

        self.frontend = cfg.frontend()
        self.module_list = torch.nn.ModuleList([ConformerRelPosBlockV1(cfg.block_cfg) for _ in range(cfg.num_layers)])



from enum import Enum
from typing import Callable, Union

from torch import nn


class BoundaryClassV1(Enum):
    """Phoneme state class augmentation selector"""

    none = 1
    word_end = 2
    boundary = 4

    def factor(self):
        return self.value

    @classmethod
    def from_flags(cls, use_word_end_classes: bool, use_boundary_classes: bool) -> "BoundaryClassV1":
        assert not (use_word_end_classes and use_boundary_classes), "cannot use both classes"

        if use_boundary_classes:
            return cls.boundary
        elif use_word_end_classes:
            return cls.word_end
        else:
            return cls.none

    @classmethod
    def from_dense_label_info(cls, li: "i6_core.mm.context_label.DenseLabelInfo") -> "BoundaryClassV1":
        return cls.from_flags(
            use_word_end_classes=li.use_word_end_classes,
            use_boundary_classes=li.use_boundary_classes,
        )


def get_center_dim(
    n_contexts: int,
    num_hmm_states_per_phone: int,
    ph_class: Union[int, BoundaryClassV1],
) -> int:
    """
    :return: number of center phonemes given the augmentation values
    """

    factor = ph_class.factor() if isinstance(ph_class, BoundaryClassV1) else ph_class
    return n_contexts * num_hmm_states_per_phone * factor


def get_mlp(
    num_input: int,
    num_output: int,
    hidden_dim: int,
    dropout: float,
    activation: Callable[[], nn.Module],
    num_layers,
) -> nn.Module:
    """
    :return: a context-mixing MLP according to the specifications
    """

    assert num_input > 0
    assert num_output > 0
    assert num_layers > 0
    assert hidden_dim > 0
    assert 0.0 <= dropout <= 1.0

    return nn.Sequential(
        *[
            layer
            for in_dim in [
                num_input,
                *[hidden_dim for _ in range(num_layers - 1)],
            ]
            for layer in [
                nn.Linear(in_dim, hidden_dim),
                activation(),
                nn.Dropout(dropout),
            ]
        ],
        nn.Linear(hidden_dim, num_output),
    )



from __future__ import annotations

__all__ = ["ConformerMHSAV1", "ConformerMHSAV1Config", "ConformerMHSAV2", "ConformerMHSAV2Config"]

from dataclasses import dataclass
from typing import Optional, Literal
import torch

from i6_models.config import ModelConfiguration
from i6_models.util import compat
from i6_models.parts.dropout import BroadcastDropout


@dataclass
class ConformerMHSAV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dim and total dimension for query/key and value projections, should be divisible by `num_att_heads`
        num_att_heads: number of attention heads
        att_weights_dropout: attention weights dropout
        dropout: multi-headed self attention output dropout
    """

    input_dim: int
    num_att_heads: int
    att_weights_dropout: float
    dropout: float

    def __post_init__(self) -> None:
        super().__post_init__()
        assert self.input_dim % self.num_att_heads == 0, "input_dim must be divisible by num_att_heads"


class ConformerMHSAV1(torch.nn.Module):
    """
    Conformer multi-headed self-attention module
    """

    def __init__(self, cfg: ConformerMHSAV1Config):
        super().__init__()

        self.layernorm = torch.nn.LayerNorm(cfg.input_dim)
        self.mhsa = torch.nn.MultiheadAttention(
            cfg.input_dim, cfg.num_att_heads, dropout=cfg.att_weights_dropout, batch_first=True
        )
        self.dropout = cfg.dropout

    def forward(self, input_tensor: torch.Tensor, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        Apply layer norm and multi-head self attention and dropout

        :param input_tensor: Input to the self attention of shape (B, T, F)
        :param sequence_mask: bool mask of shape (B, T), True signals within sequence, False outside, will be inverted to match the torch.nn.MultiheadAttention module
        which will be applied/added to dot product, used to mask padded key positions out
        """
        inv_sequence_mask = compat.logical_not(sequence_mask)
        output_tensor = self.layernorm(input_tensor)  # [B,T,F]

        output_tensor, _ = self.mhsa(
            output_tensor, output_tensor, output_tensor, key_padding_mask=inv_sequence_mask, need_weights=False
        )  # [B,T,F]
        output_tensor = torch.nn.functional.dropout(output_tensor, p=self.dropout, training=self.training)  # [B,T,F]

        return output_tensor


@dataclass
class ConformerMHSAV2Config(ConformerMHSAV1Config):
    """
    New attribute:
        dropout_broadcast_axes: string of axes to which dropout is broadcast, e.g. "T" for broadcasting to the time axis
                                setting to None to disable broadcasting
    """

    dropout_broadcast_axes: Optional[Literal["B", "T", "BT"]]

    def check_valid(self):
        assert self.dropout_broadcast_axes in [
            None,
            "B",
            "T",
            "BT",
        ], "invalid value, supported are None, 'B', 'T' and 'BT'"

    def __post__init__(self):
        super().__post_init__()
        self.check_valid()


class ConformerMHSAV2(ConformerMHSAV1):
    """
    Augments ConformerMHSAV1 with dropout broadcasting
    """

    def __init__(self, cfg: ConformerMHSAV2Config):
        super().__init__(cfg)

        self.dropout = BroadcastDropout(cfg.dropout, dropout_broadcast_axes=cfg.dropout_broadcast_axes)

    def forward(self, input_tensor: torch.Tensor, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        Apply layer norm and multi-head self attention and dropout

        :param input_tensor: Input to the self attention of shape (B, T, F)
        :param sequence_mask: Bool mask of shape (B, T), True signals within sequence, False outside, will be inverted to match the torch.nn.MultiheadAttention module
                              which will be applied/added to dot product, used to mask padded key positions out
        """
        inv_sequence_mask = compat.logical_not(sequence_mask)
        output_tensor = self.layernorm(input_tensor)  # [B,T,F]

        output_tensor, _ = self.mhsa(
            output_tensor, output_tensor, output_tensor, key_padding_mask=inv_sequence_mask, need_weights=False
        )  # [B,T,F]

        output_tensor = self.dropout(output_tensor)

        return output_tensor  # [B,T,F]



__all__ = ["LogUniformSampler"]


import torch
from torch import nn
from typing import Optional


class LogUniformSampler(nn.Module):
    def __init__(
        self, num_classes: int, *, distribution_clamp_min: float = 1e-10, device: Optional[torch.device] = None
    ):
        """
        Samples from a log uniform distribution from classes. Sampling is performed with replacement, i.e. sampled
        indices can appear more than once in sampled set. This can be implemented with
        `torch.distributions.multinomial.Multinomial` or `torch.multinomial`.

        :param num_classes: number of classes from which the distribution is sampled. The class indices are sorted in
            descending order according to their frequency.
        :param device: device on which the distribution is sampled.
        """
        super().__init__()

        # assumes count-sorted vocabulary, descending
        self.num_classes = num_classes

        # approximately zipf distribution
        ws = torch.arange(self.num_classes, dtype=torch.get_default_dtype(), device=device)
        self._distribution = (torch.log1p(ws + 1) - torch.log1p(ws)) / torch.log1p(torch.tensor(self.num_classes))
        self._distribution.clamp_(min=distribution_clamp_min)
        self._distribution /= self._distribution.sum()

        self._cat_sampler = torch.distributions.categorical.Categorical(probs=self._distribution)

    def sample(self, num_samples: int) -> torch.Tensor:
        """
        Returns a random tensor in the size of [num_samples].

        :param num_samples: number of samples.
        :return: [num_samples]
        """
        return self._cat_sampler.sample(torch.Size([num_samples]))

    def log_prob(self, indices: torch.Tensor) -> torch.Tensor:
        """
        Return log-probability of the given indices in the size of [B x T]

        :param indices: the ground truth target labels as indices.
        :return: [B x T]
        """
        return self._cat_sampler.log_prob(indices)



from __future__ import annotations

__all__ = [
    "ConformerPositionwiseFeedForwardV1",
    "ConformerPositionwiseFeedForwardV1Config",
    "ConformerPositionwiseFeedForwardV2",
    "ConformerPositionwiseFeedForwardV2Config",
]

from dataclasses import dataclass
from typing import Callable, Optional, Literal

import torch
from torch import nn

from i6_models.config import ModelConfiguration
from i6_models.parts.dropout import BroadcastDropout


@dataclass
class ConformerPositionwiseFeedForwardV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dimension
        hidden_dim: hidden dimension (normally set to 4*input_dim as suggested by the paper)
        dropout: dropout probability
        activation: activation function
    """

    input_dim: int
    hidden_dim: int
    dropout: float
    activation: Callable[[torch.Tensor], torch.Tensor] = nn.functional.silu


class ConformerPositionwiseFeedForwardV1(nn.Module):
    """
    Conformer feedforward module
    """

    def __init__(self, cfg: ConformerPositionwiseFeedForwardV1Config):
        super().__init__()

        self.layer_norm = nn.LayerNorm(cfg.input_dim)
        self.linear_ff = nn.Linear(in_features=cfg.input_dim, out_features=cfg.hidden_dim, bias=True)
        self.activation = cfg.activation
        self.linear_out = nn.Linear(in_features=cfg.hidden_dim, out_features=cfg.input_dim, bias=True)
        self.dropout = cfg.dropout

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        :param tensor: shape [B,T,F], F=input_dim
        :return: shape [B,T,F], F=input_dim
        """
        tensor = self.layer_norm(tensor)
        tensor = self.linear_ff(tensor)  # [B,T,F]
        tensor = self.activation(tensor)  # [B,T,F]
        tensor = nn.functional.dropout(tensor, p=self.dropout, training=self.training)  # [B,T,F]
        tensor = self.linear_out(tensor)  # [B,T,F]
        tensor = nn.functional.dropout(tensor, p=self.dropout, training=self.training)  # [B,T,F]
        return tensor


@dataclass
class ConformerPositionwiseFeedForwardV2Config(ModelConfiguration):
    """
    New attribute:
        dropout_broadcast_axes: string of axes to which dropout is broadcast, e.g. "T" for broadcasting to the time axis
                                setting to None to disable broadcasting
    Default value for `activation` removed
    """

    input_dim: int
    hidden_dim: int
    dropout: float
    activation: Callable[[torch.Tensor], torch.Tensor]
    dropout_broadcast_axes: Optional[Literal["B", "T", "BT"]]

    def check_valid(self):
        assert self.dropout_broadcast_axes in [
            None,
            "B",
            "T",
            "BT",
        ], "invalid value, supported are None, 'B', 'T' and 'BT'"

    def __post__init__(self):
        super().__post_init__()
        self.check_valid()


class ConformerPositionwiseFeedForwardV2(ConformerPositionwiseFeedForwardV1):
    """
    Augments ConformerPositionwiseFeedForwardV1 with dropout broadcasting
    """

    def __init__(self, cfg: ConformerPositionwiseFeedForwardV2Config):
        super().__init__(cfg)

        self.dropout = BroadcastDropout(cfg.dropout, dropout_broadcast_axes=cfg.dropout_broadcast_axes)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        :param tensor: shape [B,T,F], F=input_dim
        :return: shape [B,T,F], F=input_dim
        """
        tensor = self.layer_norm(tensor)
        tensor = self.linear_ff(tensor)  # [B,T,F]
        tensor = self.activation(tensor)  # [B,T,F]

        tensor = self.dropout(tensor)  # [B,T,F]
        tensor = self.linear_out(tensor)  # [B,T,F]
        tensor = self.dropout(tensor)  # [B,T,F]

        return tensor



__all__ = [
    "NoiseContrastiveEstimationLossV1",
]

import torch
from torch import nn
from torch.nn import functional as F
from typing import Optional


class NoiseContrastiveEstimationLossV1(nn.Module):
    __constants__ = ["num_samples", "log_norm_term", "reduction"]
    num_samples: int
    log_norm_term: float

    def __init__(
        self,
        num_samples: int,
        *,
        model_output_weights: torch.Tensor,
        model_output_bias: torch.Tensor,
        noise_distribution_sampler: nn.Module,
        log_norm_term: Optional[float] = None,
        reduction: str = "none",
        device: Optional[str] = None,
    ) -> None:
        """
        Noise contrastive estimation (NCE) loss implementation.
        Used to estimate the softmax. Normally for very large softmax sizes, for example word-level LM.

        :param num_samples: num of samples for the estimation, normally a value between 1000-4000.
            2000 is a good starting point.
        :param model_output_weights: represents the weights of the output layer of the model.
        :param model_output_bias: represents the bias of the output layer of the model.
        :param noise_distribution_sampler: for example `i6_models.parts.samplers.LogUniformSampler`.
        :param log_norm_term: normalisation term for true/sampled logits.
        :param reduction: reduction method for binary cross entropy.
        :param device: device where the loss will be placed.
        """
        super().__init__()

        self.num_samples = num_samples
        self.model_output_weights = model_output_weights
        self.model_output_bias = model_output_bias
        self.noise_distribution_sampler = noise_distribution_sampler
        self.log_norm_term = log_norm_term
        self.device = device

        self._bce = nn.BCEWithLogitsLoss(reduction=reduction)

    def forward(self, encoder_output: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        """

        :param encoder_output: here the INPUT to the final linear layer (variable `output` in the encoder model). Logit
            computation is performed in this method, requiring the raw encoder output fo be fed. The implementation
            requires a tensor where batch and time are flattened, requiring an `encoder_output` tensor shape [B x T, F].
        :param labels: ground truth labels. A tensor with flattened batch and time is required, requiring a `labels`
            tensor shape [B x T].
        :return: [B x T, 1 + num_sampled] if reduction == "none"
        """
        with torch.no_grad():
            samples = self.noise_distribution_sampler.sample(self.num_samples).cuda()

            # log-probabilities for the noise distribution k * q(w|h)
            ws = torch.log(torch.tensor(float(self.num_samples)))
            sampled_prob = ws + self.noise_distribution_sampler.log_prob(samples)  # [num_samples]
            true_sample_prob = ws + self.noise_distribution_sampler.log_prob(labels)  # [B x T]

        all_classes = torch.cat((labels, samples), 0)  # [B x T + num_sampled]

        # Steps:
        # - lookup embeddings + bias
        # - compute logits for log p(w|h) = s(w, h)
        # - compute log [p(w|h) / (k q(w|h))] = s(w, h) - log [k q(w|h)]
        # - feed into BCE with logits loss

        # do lookup once
        all_emb = F.embedding(all_classes, self.model_output_weights)  # [B x T + num_sampled, F]
        all_b = F.embedding(all_classes, torch.unsqueeze(self.model_output_bias, 1))  # [B X T + num_sampled, 1]

        # slice embeddings for targets and samples below
        true_emb = torch.narrow(all_emb, 0, 0, encoder_output.shape[0])  # [B x T, F]
        true_b = torch.narrow(all_b, 0, 0, encoder_output.shape[0])  # [B x T, 1]

        sampled_emb = torch.narrow(all_emb, 0, encoder_output.shape[0], self.num_samples)  # [num_sampled, F]
        sampled_b = torch.narrow(all_b, 0, encoder_output.shape[0], self.num_samples).squeeze(
            1
        )  # [num_sampled], remove dim for broadcasting

        # compute logits log p(w|h)
        sampled_logits = torch.matmul(encoder_output, sampled_emb.T)  # [B x T, num_sampled]

        # row-wise dot product
        true_logits = torch.multiply(encoder_output, true_emb)  # [B x T, F]
        true_logits = torch.sum(true_logits, 1, keepdim=True)  # [B x T, 1]

        true_logits += true_b
        sampled_logits += sampled_b

        # divide by optional constant normalization term here. Default is 1 (or 0 in log-space)
        if self.log_norm_term is not None:
            true_logits -= self.log_norm_term
            sampled_logits -= self.log_norm_term

        true_logits -= true_sample_prob.unsqueeze(1)
        sampled_logits -= sampled_prob.unsqueeze(0)

        out_logits = torch.cat((true_logits, sampled_logits), 1)  # [B x T, 1 + num_sampled]

        targets = torch.cat(
            (torch.ones_like(true_logits), torch.zeros_like(sampled_logits)), 1
        )  # [B x T, 1 + num_sampled]

        # no reduction on last axis here: [B x T, 1 + num_sampled]
        return self._bce(out_logits, targets)



from __future__ import annotations
from itertools import product

import torch
from torch import nn

from i6_models.parts.conformer.convolution import ConformerConvolutionV2, ConformerConvolutionV2Config
from i6_models.parts.conformer.feedforward import (
    ConformerPositionwiseFeedForwardV2,
    ConformerPositionwiseFeedForwardV2Config,
)
from i6_models.parts.conformer.mhsa_rel_pos import ConformerMHSARelPosV1Config, ConformerMHSARelPosV1
from i6_models.parts.conformer.norm import LayerNormNC


def test_ConformerConvolutionV2():
    def get_output_shape(
        batch,
        time,
        features,
        norm=None,
        kernel_size=31,
        dropout=0.1,
        activation=nn.functional.silu,
        dropout_broadcast_axes=None,
    ):
        x = torch.randn(batch, time, features)
        if norm is None:
            norm = nn.BatchNorm1d(features)
        cfg = ConformerConvolutionV2Config(
            channels=features,
            kernel_size=kernel_size,
            dropout=dropout,
            activation=activation,
            norm=norm,
            dropout_broadcast_axes=dropout_broadcast_axes,
        )
        conformer_conv_part = ConformerConvolutionV2(cfg)
        y = conformer_conv_part(x)
        return y.shape

    assert get_output_shape(10, 50, 250) == (10, 50, 250)
    assert get_output_shape(10, 50, 250, activation=nn.functional.relu) == (10, 50, 250)  # different activation
    assert get_output_shape(10, 50, 250, norm=LayerNormNC(250)) == (10, 50, 250)  # different norm
    assert get_output_shape(1, 50, 100) == (1, 50, 100)  # test with batch size 1
    assert get_output_shape(10, 1, 50) == (10, 1, 50)  # time dim 1
    assert get_output_shape(10, 10, 20, dropout=0.0) == (10, 10, 20)  # dropout 0
    assert get_output_shape(10, 10, 20, kernel_size=3) == (10, 10, 20)  # odd kernel size
    assert get_output_shape(5, 480, 512, dropout_broadcast_axes="T") == (5, 480, 512)  # dropout broadcast to T
    assert get_output_shape(5, 480, 512, dropout_broadcast_axes="BT") == (5, 480, 512)  # dropout broadcast to BT


def test_ConformerPositionwiseFeedForwardV2():
    def get_output_shape(input_shape, input_dim, hidden_dim, dropout, activation, dropout_broadcast_axes=None):
        x = torch.randn(input_shape)
        cfg = ConformerPositionwiseFeedForwardV2Config(
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            dropout=dropout,
            activation=activation,
            dropout_broadcast_axes=dropout_broadcast_axes,
        )
        conf_ffn_part = ConformerPositionwiseFeedForwardV2(cfg)
        y = conf_ffn_part(x)
        return y.shape

    for input_dim, hidden_dim, dropout, activation, dropout_broadcast_axes in product(
        [10, 20], [100, 200], [0.1, 0.3], [nn.functional.silu, nn.functional.relu], [None, "B", "T", "BT"]
    ):
        input_shape = (10, 100, input_dim)
        assert get_output_shape(input_shape, input_dim, hidden_dim, dropout, activation) == input_shape


def test_ConformerMHSARelPosV1():
    def get_output_shape(
        input_shape,
        seq_len,
        input_dim,
        with_bias=True,
        num_att_heads=8,
        att_weights_dropout=0.1,
        dropout=0.1,
        learnable_pos_emb=True,
        with_linear_pos=False,
        separate_pos_emb_per_head=False,
        rel_pos_clip=16,
        with_pos_bias=False,
        pos_emb_dropout=0.0,
        dropout_broadcast_axes=None,
    ):
        assert len(input_shape) == 3 and input_shape[-1] == input_dim

        cfg = ConformerMHSARelPosV1Config(
            input_dim=input_dim,
            num_att_heads=num_att_heads,
            with_bias=with_bias,
            att_weights_dropout=att_weights_dropout,
            dropout=dropout,
            learnable_pos_emb=learnable_pos_emb,
            with_linear_pos=with_linear_pos,
            separate_pos_emb_per_head=separate_pos_emb_per_head,
            rel_pos_clip=rel_pos_clip,
            with_pos_bias=with_pos_bias,
            pos_emb_dropout=pos_emb_dropout,
            dropout_broadcast_axes=dropout_broadcast_axes,
        )
        conf_mhsa_rel_pos = ConformerMHSARelPosV1(cfg)
        input_tensor = torch.randn(input_shape)
        sequence_mask = torch.less(torch.arange(input_shape[1])[None, :], torch.tensor(seq_len)[:, None])

        output = conf_mhsa_rel_pos(input_tensor, sequence_mask)

        return list(output.shape)

    # with key padding mask
    input_shape = [4, 15, 32]  # B,T,F
    seq_len = [15, 12, 10, 15]

    for learnable_pos_emb, with_pos_bias, pos_emb_dropout, with_linear_pos, separate_pos_emb_per_head in product(
        [True, False], [True, False], [0.0, 0.1], [True, False], [True, False]
    ):
        assert get_output_shape(
            input_shape,
            seq_len,
            32,
            learnable_pos_emb=learnable_pos_emb,
            with_pos_bias=with_pos_bias,
            pos_emb_dropout=pos_emb_dropout,
            with_linear_pos=with_linear_pos,
            separate_pos_emb_per_head=separate_pos_emb_per_head,
        ) == [4, 15, 32]


def test_ConformerMHSARelPosV1_against_Espnet():
    from espnet2.asr_transducer.encoder.modules.attention import RelPositionMultiHeadedAttention
    from espnet2.asr_transducer.encoder.modules.positional_encoding import RelPositionalEncoding

    num_heads = 4
    embed_size = 256
    dropout_rate = 0.1
    batch_dim_size = 4
    time_dim_size = 50
    seq_len = torch.Tensor([50, 10, 20, 40])
    sequence_mask = torch.less(torch.arange(time_dim_size)[None, :], seq_len[:, None])

    espnet_mhsa_module = RelPositionMultiHeadedAttention(
        num_heads=num_heads, embed_size=embed_size, dropout_rate=dropout_rate
    )
    espnet_mhsa_module.eval()
    espnet_pos_enc_module = RelPositionalEncoding(embed_size, dropout_rate=dropout_rate)
    espnet_pos_enc_module.eval()

    cfg = ConformerMHSARelPosV1Config(
        input_dim=embed_size,
        num_att_heads=num_heads,
        with_bias=True,
        att_weights_dropout=dropout_rate,
        dropout=dropout_rate,
        learnable_pos_emb=False,
        with_linear_pos=True,
        separate_pos_emb_per_head=True,
        rel_pos_clip=None,
        with_pos_bias=True,
        pos_emb_dropout=dropout_rate,
        dropout_broadcast_axes=None,
    )
    own_mhsa_module = ConformerMHSARelPosV1(cfg)
    own_mhsa_module.eval()
    own_mhsa_module.linear_pos = espnet_mhsa_module.linear_pos
    own_mhsa_module.pos_bias_u = espnet_mhsa_module.pos_bias_u
    own_mhsa_module.pos_bias_v = espnet_mhsa_module.pos_bias_v
    own_mhsa_module.out_proj = espnet_mhsa_module.linear_out
    own_mhsa_module.qkv_proj.weight = nn.Parameter(
        torch.cat(
            [
                espnet_mhsa_module.linear_q.weight,
                espnet_mhsa_module.linear_k.weight,
                espnet_mhsa_module.linear_v.weight,
            ],
            dim=0,
        )
    )
    own_mhsa_module.qkv_proj.bias = nn.Parameter(
        torch.cat(
            [espnet_mhsa_module.linear_q.bias, espnet_mhsa_module.linear_k.bias, espnet_mhsa_module.linear_v.bias],
            dim=0,
        )
    )

    input_tensor = torch.rand((batch_dim_size, time_dim_size, embed_size))
    inv_sequence_mask = torch.logical_not(sequence_mask)

    input_tensor_layernorm = own_mhsa_module.layernorm(input_tensor)

    espnet_pos_enc = espnet_pos_enc_module(input_tensor_layernorm)
    espnet_output_tensor = espnet_mhsa_module(
        query=input_tensor_layernorm,
        key=input_tensor_layernorm,
        value=input_tensor_layernorm,
        pos_enc=espnet_pos_enc,
        mask=inv_sequence_mask,
    )

    own_output_tensor = own_mhsa_module(input_tensor, sequence_mask=sequence_mask)

    assert torch.allclose(espnet_output_tensor, own_output_tensor, rtol=1e-03, atol=1e-6)



from typing import Optional, Literal

import torch
from torch import nn


class BroadcastDropout(nn.Module):
    """
    customized dropout module supporting dropout broadcasting
    supported variants are:
        - no broadcasting (default): dropout_broadcast_axes=None
        - broadcast over the batch axis: dropout_broadcast_axes='B'
        - broadcast over the time axis: dropout_broadcast_axes='T'
        - broadcast over the batch and time axes: dropout_broadcast_axes='BT'
    """

    def __init__(self, p: float, dropout_broadcast_axes: Optional[Literal["B", "T", "BT"]] = None):
        super().__init__()

        self.p = p
        assert dropout_broadcast_axes in [
            None,
            "B",
            "T",
            "BT",
        ], "invalid value, supported are None, 'B', 'T' and 'BT'"
        self.dropout_broadcast_axes = dropout_broadcast_axes

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        :param tensor: input tensor of shape [B, T, F]
        :return: tensor of shape [B, T, F]
        """
        if self.dropout_broadcast_axes is None:
            tensor = torch.nn.functional.dropout(tensor, p=self.p, training=self.training)
        elif self.dropout_broadcast_axes == "T":  # [B, T, F] -> [B, F, T] -> [B, T, F]
            # torch.nn.functional.dropout1d expects a 3D tensor and broadcasts in the last dimension.
            tensor = torch.nn.functional.dropout1d(tensor.transpose(1, 2), p=self.p, training=self.training).transpose(
                1, 2
            )
        elif self.dropout_broadcast_axes == "B":  # [B, T, F] -> [T, F, B] -> [B, T, F]
            tensor = torch.nn.functional.dropout1d(tensor.permute(1, 2, 0), p=self.p, training=self.training).permute(
                2, 0, 1
            )
        elif self.dropout_broadcast_axes == "BT":  # [B, T, F] -> [B*T, F] -> [F, B*T] -> [B*T, F] -> [B, T, F]
            batch_dim_size = tensor.shape[0]
            feature_dim_size = tensor.shape[-1]

            tensor = (
                torch.nn.functional.dropout1d(
                    tensor.reshape(-1, feature_dim_size).transpose(0, 1), p=self.p, training=self.training
                )
                .transpose(0, 1)
                .reshape(batch_dim_size, -1, feature_dim_size)
            )

        return tensor



__all__ = [
    "FactoredDiphoneBlockV1Config",
    "FactoredDiphoneBlockV1",
    "FactoredDiphoneBlockV2Config",
    "FactoredDiphoneBlockV2",
]

from dataclasses import dataclass
from typing import Callable, Tuple, Union

import torch
from torch import nn, Tensor
import torch.nn.functional as F

from i6_models.config import ModelConfiguration

from .util import BoundaryClassV1, get_center_dim, get_mlp


@dataclass
class FactoredDiphoneBlockV1Config(ModelConfiguration):
    """
    Attributes:
        num_contexts: the number of raw phonemes/acoustic contexts
        num_hmm_states_per_phone: the number of HMM states per phoneme
        boundary_class: the phoneme state augmentation to apply

        activation: activation function to use in the context mixing MLP.
        context_mix_mlp_dim: inner dimension of the context mixing MLP layers
        context_mix_mlp_num_layers: how many hidden layers on the MLPs there should be
        left_context_embedding_dim: embedding dimension of the left context
            values. Good choice is in the order of num_contexts.

        dropout: dropout probabilty
        num_inputs: input dimension of the output block, must match w/ output dimension
            of main encoder (e.g. Conformer)
    """

    num_contexts: int
    num_hmm_states_per_phone: int
    boundary_class: Union[int, BoundaryClassV1]

    activation: Callable[[], nn.Module]
    context_mix_mlp_dim: int
    context_mix_mlp_num_layers: int
    left_context_embedding_dim: int

    dropout: float
    num_inputs: int

    def __post_init__(self):
        super().__post_init__()

        assert self.num_contexts > 0
        assert self.num_hmm_states_per_phone > 0

        assert self.context_mix_mlp_dim > 0
        assert self.context_mix_mlp_num_layers > 0
        assert self.left_context_embedding_dim > 0

        assert self.num_inputs > 0
        assert 0.0 <= self.dropout <= 1.0, "dropout must be a probability"


class FactoredDiphoneBlockV1(nn.Module):
    """
    Diphone FH model output block.

    Consumes the output h(x) of a main encoder model and computes factored or joint
    output logits/probabilities for p(c|l,h(x)) and p(l|h(x)).
    """

    def __init__(self, cfg: FactoredDiphoneBlockV1Config):
        super().__init__()

        self.boundary_class = cfg.boundary_class
        self.num_contexts = cfg.num_contexts
        self.num_hmm_states_per_phone = cfg.num_hmm_states_per_phone

        self.num_center = get_center_dim(self.num_contexts, self.num_hmm_states_per_phone, self.boundary_class)
        self.num_diphone = self.num_center * self.num_contexts

        self.left_context_encoder = get_mlp(
            num_input=cfg.num_inputs,
            num_output=cfg.num_contexts,
            hidden_dim=cfg.context_mix_mlp_dim,
            num_layers=cfg.context_mix_mlp_num_layers,
            dropout=cfg.dropout,
            activation=cfg.activation,
        )
        self.left_context_embedding = nn.Embedding(cfg.num_contexts, cfg.left_context_embedding_dim)
        self.center_encoder = get_mlp(
            num_input=cfg.num_inputs + cfg.left_context_embedding_dim,
            num_output=self.num_center,
            hidden_dim=cfg.context_mix_mlp_dim,
            num_layers=cfg.context_mix_mlp_num_layers,
            dropout=cfg.dropout,
            activation=cfg.activation,
        )

    def forward(self, *args, **kwargs) -> Tuple[Tensor, Tensor, Tensor]:
        return self.forward_factored(*args, **kwargs)

    def forward_factored(
        self,
        features: Tensor,  # B, T, F
        contexts_left: Tensor,  # B, T
        **kwargs,  # kwargs because the train_step function passes contexts_center for right context training
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        :param features: Main encoder output. shape B, T, F. F=num_inputs
        :param contexts_left: The left contexts used to compute p(c|l,x), shape B, T.
            Valid values range from [0, num_contexts).
        :return: tuple of logits for p(c|l,x), p(l|x) and the embedded left context values.
        """

        logits_left = self.left_context_encoder(features)  # B, T, C
        # in training we forward exactly one context per T, so: B, T, E
        contexts_embedded_left = self.left_context_embedding(contexts_left)
        features_center = torch.cat((features, contexts_embedded_left), -1)  # B, T, F+E
        logits_center = self.center_encoder(features_center)  # B, T, C

        return logits_center, logits_left, contexts_embedded_left

    def forward_joint(self, features: Tensor) -> Tensor:
        """
        :param features: Main encoder output. shape B, T, F. F=num_inputs
        :return: log probabilities for p(c,l|x).
        """

        logits_left = self.left_context_encoder(features)  # B, T, C

        # here we forward every context to compute p(c, l|x) = p(c|l, x) * p(l|x)
        contexts_left = torch.arange(self.num_contexts, device=features.device)  # C
        contexts_embedded_left = self.left_context_embedding(contexts_left)  # C, E

        features_expanded = features.expand((self.num_contexts, -1, -1, -1))  # C, B, T, F
        contexts_embedded_left_ = contexts_embedded_left.reshape((self.num_contexts, 1, 1, -1)).expand(
            (-1, features.shape[0], features.shape[1], -1)
        )  # C, B, T, E
        features_center = torch.cat((features_expanded, contexts_embedded_left_), dim=-1)  # C, B, T, F+E
        logits_center = self.center_encoder(features_center)  # C, B, T, F'
        log_probs_center = F.log_softmax(logits_center, -1)
        log_probs_center = log_probs_center.permute((1, 2, 3, 0))  # B, T, F', C
        log_probs_left = F.log_softmax(logits_left, -1)
        log_probs_left = log_probs_left.unsqueeze(-2)  # B, T, 1, C

        joint_log_probs = log_probs_center + log_probs_left  # B, T, F', C
        joint_log_probs = joint_log_probs.reshape(
            (features.shape[0], features.shape[1], self.num_diphone)
        )  # B, T, F'*C

        return joint_log_probs


@dataclass
class FactoredDiphoneBlockV2Config(FactoredDiphoneBlockV1Config):
    """
    Attributes:
        Same attributes as parent class. In addition:

        center_state_embedding_dim: embedding dimension of the center state
            values. Good choice is in the order of num_center_states.
    """

    center_state_embedding_dim: int

    def __post_init__(self):
        super().__post_init__()

        assert self.center_state_embedding_dim > 0


class FactoredDiphoneBlockV2(FactoredDiphoneBlockV1):
    """
    Like FactoredDiphoneBlockV1, but computes an additional diphone output on the right context `p(r|c,x)`.

    This additional output is ignored when computing the joint, and only used in training.
    """

    def __init__(self, cfg: FactoredDiphoneBlockV2Config):
        super().__init__(cfg)

        self.center_state_embedding = nn.Embedding(self.num_center, cfg.center_state_embedding_dim)
        self.right_context_encoder = get_mlp(
            num_input=cfg.num_inputs + cfg.center_state_embedding_dim,
            num_output=self.num_contexts,
            hidden_dim=cfg.context_mix_mlp_dim,
            num_layers=cfg.context_mix_mlp_num_layers,
            dropout=cfg.dropout,
            activation=cfg.activation,
        )

    # update type definitions
    def forward(self, *args, **kwargs) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        return super().forward(*args, **kwargs)

    def forward_factored(
        self,
        features: Tensor,  # B, T, F
        contexts_left: Tensor,  # B, T
        contexts_center: Tensor,  # B, T
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        """
        :param features: Main encoder output. shape B, T, F. F=num_inputs
        :param contexts_left: The left contexts used to compute p(c|l,x), shape B, T.
            Valid values range from [0, num_contexts).
        :param contexts_center: The center states used to compute p(r|c,x), shape B, T.
            Given that the center state also contains the word-end class and HMM state ID, the valid values
            range from [0, num_center_states), where num_center_states >= num_contexts.
        :return: tuple of logits for p(c|l,x), p(l|x), p(r|c,x) and the embedded left context and center state values.
        """

        logits_center, logits_left, contexts_left_embedded = super().forward_factored(features, contexts_left)

        # in training we forward exactly one context per T, so: B, T, E
        center_states_embedded = self.center_state_embedding(contexts_center)
        features_right = torch.cat((features, center_states_embedded), -1)  # B, T, F+E
        logits_right = self.right_context_encoder(features_right)

        return logits_center, logits_left, logits_right, contexts_left_embedded, center_states_embedded



__all__ = [
    "LstmEncoderV1Config",
    "LstmEncoderV1",
]

from dataclasses import dataclass
import torch
from torch import nn
from typing import Any, Dict, Optional, Tuple, Union

from i6_models.config import ModelConfiguration
from i6_models.parts.lstm import LstmBlockV1Config, LstmBlockV1


@dataclass
class LstmEncoderV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dimension size
        embed_dim: embedding dimension
        embed_dropout: dropout layer after the embedding layer
        lstm_layers_cfg: configuration of the lstm block
        lstm_dropout: dropout layer after the lstm block
        init_args: used to initialize parameters of modules, example:
            ```
            {
                "init_args_w": {"func": "normal", "arg": {"mean": 0.0, "std": 0.1}},
                "init_args_b": {"func": "normal", "arg": {"mean": 0.0, "std": 0.1}},
            }
            ```
    """

    input_dim: int
    embed_dim: int
    embed_dropout: float
    lstm_layers_cfg: LstmBlockV1Config
    lstm_dropout: float
    init_args: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        if self.init_args is not None:
            for _, val in self.init_args.items():
                assert "func" in val.keys()
                assert "arg" in val.keys()

    @classmethod
    def from_dict(cls, model_cfg_dict: Dict[str, Any]):
        model_cfg_dict = model_cfg_dict.copy()
        model_cfg_dict["lstm_layers_cfg"] = LstmBlockV1Config.from_dict(model_cfg_dict["lstm_layers_cfg"])
        return cls(**model_cfg_dict)


class LstmEncoderV1(nn.Module):
    def __init__(self, model_cfg: Union[LstmEncoderV1Config, Dict[str, Any]]):
        """
        Model definition of LSTM encoder. Contains embedding layer followed by single lstm stack, dropout after both.
        Padding sequence in forward call.

        :param model_cfg: holds model configuration as dataclass or dict instance.
        """
        super().__init__()

        self.cfg = LstmEncoderV1Config.from_dict(model_cfg) if isinstance(model_cfg, dict) else model_cfg

        self.embedding = nn.Embedding(self.cfg.input_dim, self.cfg.embed_dim)
        self.embed_dropout = nn.Dropout(self.cfg.embed_dropout)

        self.lstm_block = LstmBlockV1(self.cfg.lstm_layers_cfg)
        self.lstm_dropout = nn.Dropout(self.cfg.lstm_dropout)

        if self.cfg.init_args is not None:
            self._param_init(**self.cfg.init_args)

    def _param_init(self, init_args_w=None, init_args_b=None):
        for m in self.modules():
            for name, param in m.named_parameters():
                if "bias" in name:
                    if init_args_b["func"] == "normal":
                        init_func = nn.init.normal_
                    else:
                        raise NotImplementedError
                    hyp = init_args_b["arg"]
                else:
                    if init_args_w["func"] == "normal":
                        init_func = nn.init.normal_
                    else:
                        raise NotImplementedError
                    hyp = init_args_w["arg"]
                init_func(param, **hyp)

    def forward(self, x: torch.Tensor, seq_len: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """

        :param x: [B, T, input_dim]
        :param seq_len: [B]
        :return: [B, T, lstm_layers_cfg.hidden_dim]
        """
        embed = self.embedding(x)
        embed = self.embed_dropout(embed)
        out, _ = self.lstm_block(embed, seq_len)
        out = self.lstm_dropout(out)

        return out, seq_len

    def forward_with_state(
        self,
        x: torch.Tensor,
        seq_len: torch.Tensor,
        lstm_h: Optional[torch.Tensor] = None,
        lstm_c: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        embed = self.embedding(x)
        embed = self.embed_dropout(embed)

        out, lstm_h, lstm_c = self.lstm_block.forward_with_state(embed, seq_len, lstm_h=lstm_h, lstm_c=lstm_c)
        out = self.lstm_dropout(out)

        return out, lstm_h, lstm_c



__all__ = [
    "FactoredDiphoneBlockV1Config",
    "FactoredDiphoneBlockV1",
    "FactoredDiphoneBlockV2Config",
    "FactoredDiphoneBlockV2",
    "BoundaryClassV1",
]

from .diphone import *
from .util import BoundaryClassV1



__all__ = ["LstmBlockV1Config", "LstmBlockV1"]

from dataclasses import dataclass
import torch
from torch import nn
from typing import Any, Dict, Tuple, Union, Optional

from i6_models.config import ModelConfiguration


@dataclass
class LstmBlockV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dimension size
        hidden_dim: hidden dimension of one direction of LSTM
        num_layers: number of uni-directional LSTM layers, minimum 2
        bias: add a bias term to the LSTM layer
        dropout: nn.LSTM supports internal Dropout applied between each layer of LSTM (but not on input/output)
        enforce_sorted:
            True: expects that sequences are sorted by sequence length in decreasing order.
                Will not do any sorting.
                This is required for ONNX-Export, and thus the recommended setting.
            False: no expectation.
                It will internally enforce that they are sorted
                and undo the reordering at the output.

            Sorting can for example be performed independent of the ONNX export in e.g. train_step:

                audio_features_len, indices = torch.sort(audio_features_len, descending=True)
                audio_features = audio_features[indices, :, :]
                labels = labels[indices, :]
                labels_len = labels_len[indices]
    """

    input_dim: int
    hidden_dim: int
    num_layers: int
    bias: bool
    dropout: float
    enforce_sorted: bool

    @classmethod
    def from_dict(cls, model_cfg_dict: Dict[str, Any]):
        model_cfg_dict = model_cfg_dict.copy()
        return cls(**model_cfg_dict)


class LstmBlockV1(nn.Module):
    def __init__(self, model_cfg: Union[LstmBlockV1Config, Dict[str, Any]]):
        """
        Model definition of LSTM block. Contains single lstm stack and padding sequence in forward call. Including
        dropout, batch-first variant, hardcoded to use B,T,F input.

        Supports: TorchScript, ONNX-export.

        :param model_cfg: holds model configuration as dataclass or dict instance.
        """
        super().__init__()

        self.cfg = LstmBlockV1Config.from_dict(model_cfg) if isinstance(model_cfg, dict) else model_cfg

        self.dropout = self.cfg.dropout
        self.enforce_sorted = self.cfg.enforce_sorted
        self.lstm_stack = nn.LSTM(
            input_size=self.cfg.input_dim,
            hidden_size=self.cfg.hidden_dim,
            num_layers=self.cfg.num_layers,
            bias=self.cfg.bias,
            dropout=self.dropout,
            batch_first=True,
            bidirectional=False,
        )

    def forward(self, x: torch.Tensor, seq_len: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        :param x: [B, T, input_dim]
        :param seq_len:[B], should be on CPU for Script/Trace mode
        :return: [B, T, hidden_dim]
        """
        lstm_packed_in = nn.utils.rnn.pack_padded_sequence(
            input=x, lengths=seq_len, enforce_sorted=self.enforce_sorted, batch_first=True
        )

        lstm_out, _ = self.lstm_stack(lstm_packed_in)
        lstm_out, _ = nn.utils.rnn.pad_packed_sequence(lstm_out, padding_value=0.0, batch_first=True)

        return lstm_out, seq_len

    def forward_with_state(
        self,
        x: torch.Tensor,
        seq_len: torch.Tensor,
        lstm_h: Optional[torch.Tensor] = None,
        lstm_c: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if lstm_h is None:
            lstm_packed_in = nn.utils.rnn.pack_padded_sequence(
                input=x, lengths=seq_len, enforce_sorted=self.enforce_sorted, batch_first=True
            )
            assert lstm_c is None
            lstm_out, (h_x, c_x) = self.lstm_stack(lstm_packed_in)
            lstm_out, _ = nn.utils.rnn.pad_packed_sequence(lstm_out, padding_value=0.0, batch_first=True)
        else:
            assert lstm_c is not None
            lstm_out, (h_x, c_x) = self.lstm_stack(x, (lstm_h, lstm_c))

        return lstm_out, h_x, c_x



from typing import Optional, Tuple

import torch
import torch.nn as nn
import numpy as np

__all__ = ["RandomMask"]


class RandomMask(nn.Module):
    """
    randomly mask out consecutive frames time dimension, the masked frames can be either
    replaced with zeros or with learnable embeddings.
    simplified version from Fairseq compute_mask_indices function,
    C.f. https://github.com/facebookresearch/fairseq/blob/ecbf110e1eb43861214b05fa001eff584954f65a/fairseq/data/data_utils.py#L399
    """

    def __init__(
        self,
        input_dim: int,
        mask_replace_val: str,
        mask_percentage: float,
        mask_length: int,
    ):
        """
        :param input_dim: number of feature dimension of input
        :param mask_replace_val: the way to replace masked frames, either with zeros or lernable embeddings
        :param mask_percentage: percentage of frames to be masked out
        :param mask_length: the length of each mask span
        """
        super().__init__()

        assert mask_replace_val in ["lernable", "zero"], "not implemented yet"
        if mask_replace_val == "lernable":
            self.mask_emb = nn.Parameter(torch.FloatTensor(input_dim).uniform_())
        elif mask_replace_val == "zero":
            self.mask_emb = torch.zeros(input_dim)
        self.mask_percentage = mask_percentage
        self.mask_length = mask_length

    def forward(
        self,
        tensor: torch.tensor,
        padding_mask: Optional[torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        ndim_batch, ndim_time, _ = tensor.size()

        mask = torch.zeros((ndim_batch, ndim_time), dtype=torch.bool)

        mask_idcs = []
        for i in range(ndim_batch):
            if padding_mask is not None:
                seq_len = ndim_time - padding_mask[i].long().sum().item()
                assert seq_len >= 0
            else:
                seq_len = ndim_time

            num_mask = int(
                # add a random number for probabilistic rounding
                self.mask_percentage * seq_len / float(self.mask_length) + np.random.rand()
            )

            min_len = self.mask_length
            if seq_len - min_len <= num_mask:
                min_len = seq_len - num_mask - 1
            mask_idc = np.random.choice(seq_len - min_len, num_mask, replace=False)

            for j in mask_idc:
                mask[i, j : j + self.mask_length] = True

        tensor[mask] = self.mask_emb.to(tensor.device)

        return tensor, torch.tensor(mask).to(tensor.device)



from __future__ import annotations

__all__ = ["ConformerMHSARelPosV1", "ConformerMHSARelPosV1Config"]

from dataclasses import dataclass
import math
from typing import Optional, Literal

import torch
from torch import nn
import torch.nn.functional as F

from i6_models.config import ModelConfiguration
from i6_models.util import compat
from i6_models.parts.dropout import BroadcastDropout


@dataclass
class ConformerMHSARelPosV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dim and total dimension for query/key and value projections, should be divisible by `num_att_heads`
        num_att_heads: number of attention heads
        with_bias: whether to add bias to qkv and output linear projections
        att_weights_dropout: attention weights dropout
        learnable_pos_emb: whether to use learnable relative positional embeddings instead of fixed sinusoidal ones
        rel_pos_clip: maximal relative postion for embedding
        with_linear_pos: whether to linearly transform the positional embeddings
        separate_pos_emb_per_head: whether to create head-dependent positional embeddings
        with_pos_bias: whether to add additional position bias terms to the attention scores
        pos_emb_dropout: dropout for the positional embeddings
        dropout: multi-headed self attention output dropout
        dropout_broadcast_axes: string of axes to which dropout is broadcast, e.g. "T" for broadcasting to the time axis
                                setting to None to disable broadcasting
    """

    input_dim: int
    num_att_heads: int
    with_bias: bool
    att_weights_dropout: float
    learnable_pos_emb: bool
    rel_pos_clip: Optional[int]
    with_linear_pos: bool
    with_pos_bias: bool
    separate_pos_emb_per_head: bool
    pos_emb_dropout: float
    dropout: float
    dropout_broadcast_axes: Optional[Literal["B", "T", "BT"]]

    def __post_init__(self) -> None:
        super().__post_init__()
        assert self.input_dim % self.num_att_heads == 0, "input_dim must be divisible by num_att_heads"
        assert self.dropout_broadcast_axes in [
            None,
            "B",
            "T",
            "BT",
        ], "invalid value, supported are None, 'B', 'T' and 'BT'"


class ConformerMHSARelPosV1(nn.Module):
    """
    Conformer multi-headed self-attention module supporting
        - self-attention with relative positional encoding proposed by Shaw et al. (cf. https://arxiv.org/abs/1803.02155)
            * learnable_pos_emb = True
            * with_pos_bias = False
            * with_linear_pos = False
            * separate_pos_emb_per_head = False (RETURNN default)
            * with_bias = False (RETURNN default)
        - and self-attention with Transformer-XL style relative PE by Dai et al.
            (cf. https://arxiv.org/abs/1901.02860, https://github.com/kimiyoung/transformer-xl/blob/master/pytorch/mem_transformer.py,
                 https://github.com/espnet/espnet/blob/master/espnet2/asr_transducer/encoder/modules/attention.py#L9)
            * learnable_pos_emb = False
            * with_pos_bias = True
            * with_linear_pos = False (paper implementation) / with_linear_pos = True (ESPnet default)
            * separate_pos_emb_per_head = False (paper implementation) / separate_pos_emb_per_head = True (ESPnet default)
            * with_bias = False (paper implementation) / with_bias = True (ESPnet default)
    """

    def __init__(self, cfg: ConformerMHSARelPosV1Config):
        super().__init__()

        self.layernorm = nn.LayerNorm(cfg.input_dim)

        self.embed_dim = cfg.input_dim
        self.num_heads = cfg.num_att_heads
        self.embed_dim_per_head = self.embed_dim // self.num_heads

        self.learnable_pos_emb = cfg.learnable_pos_emb
        self.rel_pos_clip = cfg.rel_pos_clip
        self.separate_pos_emb_per_head = cfg.separate_pos_emb_per_head
        self.with_pos_bias = cfg.with_pos_bias
        self.pos_emb_dropout = nn.Dropout(cfg.pos_emb_dropout)

        assert not self.learnable_pos_emb or self.rel_pos_clip

        self.att_weights_dropout = nn.Dropout(cfg.att_weights_dropout)

        assert self.embed_dim % self.num_heads == 0, "embed_dim must be divisible by num_heads"

        # projection matrices
        self.qkv_proj = nn.Linear(self.embed_dim, 3 * self.embed_dim, bias=cfg.with_bias)
        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=cfg.with_bias)

        self.register_parameter("rel_pos_embeddings", None)
        self.register_parameter("pos_bias_u", None)
        self.register_parameter("pos_bias_v", None)

        self.pos_emb_dim = (
            self.embed_dim if cfg.with_linear_pos or cfg.separate_pos_emb_per_head else self.embed_dim_per_head
        )
        if self.learnable_pos_emb:
            self.rel_pos_embeddings = nn.parameter.Parameter(torch.empty(self.rel_pos_clip * 2 + 1, self.pos_emb_dim))
        if cfg.with_linear_pos:
            self.linear_pos = nn.Linear(
                self.pos_emb_dim,
                self.embed_dim if cfg.separate_pos_emb_per_head else self.embed_dim_per_head,
                bias=False,
            )
        else:
            self.linear_pos = nn.Identity()

        if self.with_pos_bias:
            self.pos_bias_u = nn.parameter.Parameter(torch.empty(self.num_heads, self.embed_dim_per_head))
            self.pos_bias_v = nn.parameter.Parameter(torch.empty(self.num_heads, self.embed_dim_per_head))

        self.dropout = BroadcastDropout(cfg.dropout, dropout_broadcast_axes=cfg.dropout_broadcast_axes)

        self._reset_parameters()

    def _reset_parameters(self):
        if self.learnable_pos_emb:
            nn.init.xavier_normal_(self.rel_pos_embeddings)
        if self.with_pos_bias:
            # init taken from espnet default
            nn.init.xavier_uniform_(self.pos_bias_u)
            nn.init.xavier_uniform_(self.pos_bias_v)

    def forward(self, input_tensor: torch.Tensor, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        Apply layer norm and multi-head self attention and dropout

        :param input_tensor: Input to the self attention of shape (B, T, F)
        :param sequence_mask: bool mask of shape (B, T), True signals within sequence, False outside
        """
        output_tensor = self.layernorm(input_tensor)  # [B, T, F]

        time_dim_size = output_tensor.shape[1]
        batch_dim_size = output_tensor.shape[0]

        # attention mask
        # T: query seq. length, T' key/value seg length; T = T' if same input tensor
        inv_sequence_mask = compat.logical_not(sequence_mask)  # [B, T']
        mask = (
            torch.zeros_like(inv_sequence_mask, dtype=input_tensor.dtype)
            .masked_fill(inv_sequence_mask, float("-inf"))
            .reshape(batch_dim_size, 1, 1, time_dim_size)
        )  # [B, 1, 1, T']

        # query, key and value sequences
        query_seq, key_seq, value_seq = self.qkv_proj(output_tensor).chunk(3, dim=-1)  # [B, T, #heads * F']
        q = query_seq.view(batch_dim_size, -1, self.num_heads, self.embed_dim_per_head)  # [B, T, #heads, F']
        k = key_seq.view(batch_dim_size, -1, self.num_heads, self.embed_dim_per_head)  # [B, T', #heads, F']

        if self.learnable_pos_emb:
            pos_seq_q = torch.arange(time_dim_size, device=input_tensor.device)
            pos_seq_k = torch.arange(time_dim_size, device=input_tensor.device)

            distance_mat = pos_seq_k[None, :] - pos_seq_q[:, None]
            distance_mat_clipped = torch.clamp(distance_mat, -self.rel_pos_clip, self.rel_pos_clip)

            final_mat = distance_mat_clipped + self.rel_pos_clip

            rel_pos_embeddings = self.rel_pos_embeddings[final_mat]  # [T, T', pos_emb_dim]
        else:
            rel_pos_embeddings = self._sinusoidal_pe(
                torch.arange(time_dim_size - 1, -time_dim_size, -1, device=input_tensor.device, dtype=torch.float32),
                self.pos_emb_dim,
            ).view(1, 2 * time_dim_size - 1, self.pos_emb_dim)  # [1, T+T'-1, pos_emb_dim]

        # dropout relative positional embeddings
        rel_pos_embeddings = self.pos_emb_dropout(
            rel_pos_embeddings
        )  # [T, T', pos_emb_dim] or [1, T+T'-1, pos_emb_dim]
        rel_pos_embeddings = rel_pos_embeddings.unsqueeze(2)  # [T, T', 1, pos_emb_dim] or [1, T+T'-1, 1, pos_emb_dim]

        # linear transformation or identity
        rel_pos_embeddings = self.linear_pos(rel_pos_embeddings)  # [T, T', 1, F'|F] or [1, T+T'-1, 1, F'|F]

        if self.separate_pos_emb_per_head:
            rel_pos_embeddings = rel_pos_embeddings.squeeze(2).reshape(
                *rel_pos_embeddings.shape[:2], -1, self.embed_dim_per_head
            )  # [T, T', #heads, F'] or [1, T+T'-1, #heads, F']

        q_with_bias_u = q + self.pos_bias_u if self.with_pos_bias else q  # [B, T, #heads, F']
        q_with_bias_v = q + self.pos_bias_v if self.with_pos_bias else q

        # attention matrix a and c
        attn_ac = torch.einsum("bihf, bjhf -> bhij", q_with_bias_u, k)  # [B, #heads, T, T']

        # attention matrix b and d
        attn_bd = torch.einsum(
            "bihf, ijhf -> bhij", q_with_bias_v, rel_pos_embeddings
        )  # [B, #heads, T, T'] or [B, #heads, T, T+T'+1]

        if not self.learnable_pos_emb:
            attn_bd = self._rel_shift_bhij(attn_bd, k_len=time_dim_size)  # [B, #heads, T, T']

        attn = attn_ac + attn_bd + mask  # [B, #heads, T, T']
        attn_scaled = attn * (math.sqrt(1.0 / float(self.embed_dim_per_head)))  # [B, #heads, T, T']

        # softmax and dropout
        attn_output_weights = self.att_weights_dropout(F.softmax(attn_scaled, dim=-1))  # [B, #heads, T, T']

        # sequence of weighted sums over value sequence
        v = value_seq.view(batch_dim_size, -1, self.num_heads, self.embed_dim_per_head)  # [B, T, H, F']
        attn_output = torch.einsum("bhij, bjhf -> bihf", attn_output_weights, v).reshape(
            batch_dim_size, -1, self.embed_dim
        )

        output_tensor = self.out_proj(attn_output)

        output_tensor = self.dropout(output_tensor)

        return output_tensor  # [B,T,F]

    @staticmethod
    def _rel_shift_bhij(x, k_len=None):
        """
        :param x: input tensor of shape (B, H, T, L) to apply left shift
        :k_len: length of the key squence
        """
        x_shape = x.shape

        x = torch.nn.functional.pad(x, (1, 0))  # [B, H, T, L+1]
        x = x.reshape(x_shape[0], x_shape[1], x_shape[3] + 1, x_shape[2])  # [B, H, L+1, T]
        x = x[:, :, 1:]  # [B, H, L, T]
        x = x.reshape(x_shape)  # [B, H, T, L]]

        return x[:, :, :, :k_len] if k_len else x  # [B, H, T, T']

    @staticmethod
    def _sinusoidal_pe(pos_seq: torch.Tensor, embed_dim: int):
        """
        :param pos_seq: 1-D position sequence for which to compute embeddings
        :param embed_dim: embedding dimension
        """
        inv_freq = 1 / (10000 ** (torch.arange(0.0, embed_dim, 2.0, device=pos_seq.device) / embed_dim))

        sinusoid_input = torch.outer(pos_seq, inv_freq)

        pos_emb = torch.zeros(pos_seq.shape[0], embed_dim, device=pos_seq.device)

        pos_emb[:, 0::2] = sinusoid_input.sin()
        pos_emb[:, 1::2] = sinusoid_input.cos()

        return pos_emb



from itertools import product

import torch
import torch.nn as nn

from i6_models.parts.factored_hybrid import (
    BoundaryClassV1,
    FactoredDiphoneBlockV1,
    FactoredDiphoneBlockV1Config,
    FactoredDiphoneBlockV2,
    FactoredDiphoneBlockV2Config,
)
from i6_models.parts.factored_hybrid.util import get_center_dim


def test_dim_calcs():
    n_ctx = 42

    assert get_center_dim(n_ctx, 1, BoundaryClassV1.none) == 42
    assert get_center_dim(n_ctx, 1, BoundaryClassV1.word_end) == 84
    assert get_center_dim(n_ctx, 3, BoundaryClassV1.word_end) == 252
    assert get_center_dim(n_ctx, 3, BoundaryClassV1.boundary) == 504


def test_v1_output_shape_and_norm():
    n_ctx = 42
    n_in = 32

    for we_class, states_per_ph in product(
        [BoundaryClassV1.none, BoundaryClassV1.word_end, BoundaryClassV1.boundary],
        [1, 3],
    ):
        block = FactoredDiphoneBlockV1(
            FactoredDiphoneBlockV1Config(
                activation=nn.ReLU,
                context_mix_mlp_dim=64,
                context_mix_mlp_num_layers=2,
                dropout=0.1,
                left_context_embedding_dim=32,
                num_contexts=n_ctx,
                num_hmm_states_per_phone=states_per_ph,
                num_inputs=n_in,
                boundary_class=we_class,
            )
        )

        for b, t in product([10, 50, 100], [10, 50, 100]):
            contexts_forward = torch.randint(0, n_ctx, (b, t))
            encoder_output = torch.rand((b, t, n_in))
            output_center, output_left, _ = block(features=encoder_output, contexts_left=contexts_forward)
            assert output_left.shape == (b, t, n_ctx)
            cdim = get_center_dim(n_ctx, states_per_ph, we_class)
            assert output_center.shape == (b, t, cdim)

            encoder_output = torch.rand((b, t, n_in))
            output = block.forward_joint(features=encoder_output)
            cdim = get_center_dim(n_ctx, states_per_ph, we_class)
            assert output.shape == (b, t, cdim * n_ctx)
            output_p = torch.exp(output)
            ones_hopefully = torch.sum(output_p, dim=-1)
            close_to_one = torch.abs(1 - ones_hopefully).flatten() < 1e-3
            assert all(close_to_one)


def test_v2_output_shape_and_norm():
    n_ctx = 42
    n_in = 32

    for we_class, states_per_ph in product(
        [BoundaryClassV1.none, BoundaryClassV1.word_end, BoundaryClassV1.boundary],
        [1, 3],
    ):
        block = FactoredDiphoneBlockV2(
            FactoredDiphoneBlockV2Config(
                activation=nn.ReLU,
                context_mix_mlp_dim=64,
                context_mix_mlp_num_layers=2,
                dropout=0.1,
                left_context_embedding_dim=32,
                center_state_embedding_dim=128,
                num_contexts=n_ctx,
                num_hmm_states_per_phone=states_per_ph,
                num_inputs=n_in,
                boundary_class=we_class,
            )
        )

        for b, t in product([10, 50, 100], [10, 50, 100]):
            contexts_left = torch.randint(0, n_ctx, (b, t))
            contexts_center = torch.randint(0, block.num_center, (b, t))
            encoder_output = torch.rand((b, t, n_in))
            output_center, output_left, output_right, _, _ = block(
                features=encoder_output, contexts_left=contexts_left, contexts_center=contexts_center
            )
            assert output_left.shape == (b, t, n_ctx)
            assert output_right.shape == (b, t, n_ctx)
            cdim = get_center_dim(n_ctx, states_per_ph, we_class)
            assert output_center.shape == (b, t, cdim)
